import 'package:flutter/material.dart';
import 'package:provider/provider.dart';

import 'package:verifai_sdk/verifai_sdk.dart';

import 'licence.dart';
import 'modeldata.dart';

void main() {
  WidgetsFlutterBinding.ensureInitialized();
  runApp(MultiProvider(
    providers: [
      ChangeNotifierProvider(
        create: (context) => CoreResultImpl(),
      ),
      ChangeNotifierProvider(
        create: (context) => NfcResultImpl(),
      ),
      ChangeNotifierProvider(
        create: (context) => LivenessResultImpl(),
      ),
    ],
    child: const MyApp(),
  ));
}

class MyApp extends StatelessWidget {
  const MyApp({Key? key}) : super(key: key);

  @override
  Widget build(BuildContext context) {
    return MaterialApp(
      title: 'Flutter Demo',
      theme: ThemeData(
          primarySwatch: const MaterialColor(
        0xffff576d, // 0%
        <int, Color>{
          50: Color(0xffe64e62), //10%
          100: Color(0xffcc4657), //20%
          200: Color(0xffb33d4c), //30%
          300: Color(0xff993441), //40%
          400: Color(0xff802c37), //50%
          500: Color(0xff66232c), //60%
          600: Color(0xff4c1a21), //70%
          700: Color(0xff331116), //80%
          800: Color(0xff19090b), //90%
          900: Color(0xff000000), //100%
        },
      )),
      home: const MyHomePage(title: 'Verifai Flutter test page'),
    );
  }
}

class MyHomePage extends StatefulWidget {
  const MyHomePage({Key? key, required this.title}) : super(key: key);

  final String title;

  @override
  State<MyHomePage> createState() => _MyHomePageState();
}

class _MyHomePageState extends State<MyHomePage> {
  void _startCore() async {
    var api = CoreApi();
    await api.setLicence(verifaiLicence);
    api.configure(Configuration(
        enablePostCropping: true,
        enableVisualInspection: true,
        isScanHelpEnabled: false,
        enableManual: true,
    ));
    await api.start();
    Navigator.push(
      context,
      MaterialPageRoute(builder: (context) => const CoreResult()),
    );
  }

  void _startLiveness() async {
    var api = LivenessApi();
    var checks = [
      LivenessCheck(
          type: LivenessCheckType.closeEyes, args: {"numberOfSeconds": 3}),
      LivenessCheck(
          type: LivenessCheckType.tilt, args: {"faceAngleRequirement": 5}),
    ];
    var image = context.read<CoreResultImpl>().data.frontImage;
    if (image != null) {
      checks.add(LivenessCheck(
          type: LivenessCheckType.faceMatching,
          args: {"documentImage": image}));
    }
    api.start(LivenessConfiguration(showDismissButton: false, checks: checks));
  }

  void _startNfc() async {
    var api = NfcApi();
    var coreData = context.read<CoreResultImpl>().data;
    if (coreData.mrzData != null) {
      await api.start(NfcConfiguration(retrieveImage: true), coreData);
      Navigator.push(
        context,
        MaterialPageRoute(builder: (context) => const NfcResult()),
      );
    } else {
      print("No core mrz data");
    }
  }

  final verifaiStyle = ButtonStyle(
      backgroundColor:
          MaterialStateProperty.all<Color>(const Color(0xffff576d)));

  @override
  Widget build(BuildContext context) {
    return Scaffold(
      appBar: AppBar(
        title: Text(widget.title),
      ),
      body: Center(
        child: Column(
          mainAxisAlignment: MainAxisAlignment.center,
          crossAxisAlignment: CrossAxisAlignment.center,
          children: <Widget>[
            ElevatedButton(
              onPressed: _startCore,
              child: const Text('Core'),
              style: verifaiStyle,
            ),
            ElevatedButton(
              onPressed: _startNfc,
              child: const Text('NFC'),
              style: verifaiStyle,
            ),
            Consumer<LivenessResultImpl>(builder: (context, result, child) {
              if (result.data.resultList != null) {
                for (var res in result.data.resultList!) {
                  print(res?.name);
                  print(res?.checkSpecificData?["confidence"]);
                }
              }
              return ElevatedButton(
                onPressed: _startLiveness,
                child: const Text('Liveness'),
                style: verifaiStyle,
              );
            }),
          ],
        ),
      ),
    );
  }
}

class CoreResult extends StatelessWidget {
  const CoreResult({Key? key}) : super(key: key);

  @override
  Widget build(BuildContext context) {
    return Scaffold(
      appBar: AppBar(
        title: const Text('Core result'),
      ),
      body: Center(
        child: Consumer<CoreResultImpl>(builder: (context, result, child) {
          if (result.data.frontImage != null) {
            return Image.memory(result.data.frontImage!);
          } else {
            return const Text("No image available");
          }
        }),
      ),
    );
  }
}

class NfcResult extends StatelessWidget {
  const NfcResult({Key? key}) : super(key: key);

  @override
  Widget build(BuildContext context) {
    return Scaffold(
      appBar: AppBar(
        title: const Text('NFC result'),
      ),
      body: Center(
        child: Consumer<NfcResultImpl>(builder: (context, result, child) {
          if (result.data.photo != null) {
            return Image.memory(result.data.photo!);
          } else {
            return const Text("No nfc image available");
          }
        }),
      ),
    );
  }
}
