//
//  Core.swift
//  Pods-Runner
//
//  Created by Richard Chirino on 22/03/2022.
//

import Foundation
import VerifaiCommonsKit
import VerifaiKit

/// Class that handles the communication between the Flutter plugin and the Verifai SDK's Core module
class Core {
    // Define shared result object that can be used by other modules
    static var sharedResult: VerifaiResult?
    
    // Set the licence
    static func setLicence(_ licence: String, resultHandler: CoreResultApi?) {
        switch VerifaiCommons.setLicence(licence) {
        case .success(_):
            print("Successfully configured Verifai")
        case .failure(let error):
#if DEBUG
            print("🚫 Licence error: \(error)")
#endif
            resultHandler?.onErrorMessage("🚫 Licence error: \(error)", completion: { _ in })
        }
    }
    
    // Process the configuration
    static func processConfiguration(_ configuration: Configuration, resultHandler: CoreResultApi?) {
        do {
            let verifaiConfiguration = try verifaiConfiguration(from: configuration)
            try Verifai.configure(with: verifaiConfiguration)
        } catch {
            resultHandler?.onErrorMessage("🚫 Configuration error: \(error)", completion: { _ in })
        }
    }
    
    // Start the Core
    static func start(rootViewController: FlutterViewController?, resultHandler: CoreResultApi?) {
        do {
            guard let flutterVC = rootViewController else {
                resultHandler?.onErrorMessage("🚫 No root view controller found.", completion: { _ in })
                return
            }
            try Verifai.start(over: flutterVC) { result in
                switch result {
                case .failure(let error):
#if DEBUG
                    print("Error or cancellation: \(error)")
#endif
                    resultHandler?.onErrorMessage("🚫 Error or cancellation: \(error)", completion: { _ in })
                case .success(let successResult):
                    do {
                        // Save the shared result
                        sharedResult = successResult
                        // Return the result
                        let result = try makeCoreResult(from: successResult)
                        resultHandler?.onSuccessResult(result, completion: { _ in })
                    } catch {
                        resultHandler?.onErrorMessage("🚫 Error creating result: \(error)", completion: { _ in })
                    }
                }
            }
        } catch {
            print("🚫 Unhandled error: \(error)")
        }
    }
    
    // MARK: - Configuration
    
    /// Create a Verifai configuration from a flutter configuration object
    /// - Parameter fConf: The configuration object coming from flutter
    /// - Returns: A Verifai configuration object
    private static func verifaiConfiguration(from fConf: Configuration) throws -> VerifaiConfiguration {
        // Setup the validator array
        let validators: [VerifaiValidator] = verifaiValidators(from: fConf.validators)
        // Create the configuration object which we will then fill with the objects from Flutter
        let vConf = VerifaiConfiguration(requireDocumentCopy: fConf.requireDocumentCopy?.boolValue ?? true,
                                         requireCroppedImage: fConf.requireCroppedImage?.boolValue ?? true,
                                         enablePostCropping: fConf.enablePostCropping?.boolValue ?? true,
                                         enableManual: fConf.enableManual?.boolValue ?? true,
                                         requireMRZContents: fConf.requireMrzContents?.boolValue ?? false,
                                         readMRZContents: fConf.readMrzContents?.boolValue ?? true,
                                         requireNFCWhenAvailable: fConf.requireNfcWhenAvailable?.boolValue ?? false,
                                         enableVisualInspection: fConf.enableVisualInspection?.boolValue ?? false,
                                         validators: validators)
        vConf.documentFilters = verifaiDocumentFilters(from: fConf.documentFilters)
        vConf.documentFiltersAutoCreateValidators = fConf.documentFiltersAutoCreateValidators?.boolValue ?? true
        vConf.customDismissButtonTitle = fConf.customDismissButtonTitle
        vConf.instructionScreenConfiguration = try verifaiInstructionScreenConfiguration(from: fConf.instructionScreenConfiguration)
        vConf.scanHelpConfiguration = verifaiScanHelpConfiguration(from: fConf.scanHelpConfiguration)
        return vConf
    }
    
    /// Create a Verifai validator object array from the Flutter based validator objects
    /// - Parameter validators: The validators coming from Flutter
    /// - Returns: An array of Verifai based validators
    private static func verifaiValidators(from validators: [Validator]?) -> [VerifaiValidator] {
        guard let validators = validators else { return [] }
        return validators.compactMap { validator -> VerifaiValidator? in
            // Determine the type and build it if the information is correct, otherwise it will not
            // be added to the list
            switch validator.type {
            case .countryAllowlist:
                if let countryList = validator.args?["countryList"] as? [String] {
                    return VerifaiDocumentCountryWhiteListValidator(countries: countryList)
                }
            case .countryBlocklist:
                if let countryList = validator.args?["countryList"] as? [String] {
                    return VerifaiDocumentCountryBlackListValidator(countries: countryList)
                }
            case .hasMrz:
                return VerifaiDocumentHasMrzValidator()
            case .documentTypes:
                if let documentTypes = validator.args?["validDocumentTypes"] as? [Int] {
                    let verifaiDocumentTypes = VerifaiDocumentType.fromArray(documentTypes)
                    return VerifaiDocumentTypesValidator(validDocumentTypes: verifaiDocumentTypes)
                }
            case .mrzAvailable:
                return VerifaiMrzAvailableValidator()
            case .nfcKeyWhenAvailable:
                return VerifaiNFCKeyWhenAvailableValidator()
            @unknown default:
                return nil
            }
            return nil
        }
    }
    
    /// Map the document filters coming from Flutter to document filters that Verifai can understand
    /// - Parameter documentFilters: The document filters coming from Flutter
    /// - Returns: An array of Verifai based document filters
    private static func verifaiDocumentFilters(from documentFilters: [DocumentFilter]?) -> [VerifaiDocumentFilter] {
        guard let documentFilters = documentFilters else { return [] }
        return documentFilters.compactMap { filter in
            switch filter.type {
            case .documentTypeAllowList:
                if let documentTypes = filter.args?["validDocumentTypes"] as? [Int] {
                    let verifaiDocumentTypes = VerifaiDocumentType.fromArray(documentTypes)
                    return VerifaiDocumentTypeWhiteListFilter(validDocumentTypes: verifaiDocumentTypes)
                }
            case .documentAllowList:
                if let countryList = filter.args?["countryList"] as? [String] {
                    return VerifaiDocumentWhiteListFilter(countryCodes: countryList)
                }
            case .documentBlockList:
                if let countryList = filter.args?["countryList"] as? [String] {
                    return VerifaiDocumentBlackListFilter(countryCodes: countryList)
                }
            @unknown default:
                return nil
            }
            return nil
        }
    }
    
    /// Map the Flutter version of the instruction screen configuration object to Verifai's native one
    /// - Parameter configuration: The instruction screen configuration coming from flutter
    /// - Returns: Verifai's version of the instruction screen configuration
    private static func verifaiInstructionScreenConfiguration(from configuration: InstructionScreenConfiguration?) throws -> VerifaiInstructionScreenConfiguration {
        // Ensure a configuration was supplied otherwise return default
        guard let configuration = configuration else {
            return try VerifaiInstructionScreenConfiguration(showInstructionScreens: true)
        }
        // Declare a holder for the native results
        var instructionScreens: [VerifaiInstructionScreen: VerifaiSingleInstructionScreenConfiguration] = [:]
        // Get the global values first
        let showInstructionScreens = configuration.showInstructionScreens?.boolValue ?? true
        try configuration.instructionScreens?.forEach({ instructionScreen in
            // Get the correct enum
            let instructionScreenEnum = VerifaiInstructionScreen.from(instructionScreen.screen)
            // Declare holder for configuration
            var screenConfiguration: VerifaiSingleInstructionScreenConfiguration?
            // Build the configuration according to the type
            switch instructionScreen.type {
            case .media:
                // Get the values
                let instruction = NSAttributedString(string: instructionScreen.instruction ?? "")
                let values = try VerifaiInstructionScreenValues(title: instructionScreen.title ?? "",
                                                                header: instructionScreen.header ?? "",
                                                                mp4FileName: instructionScreen.mp4FileName ?? "",
                                                                instruction: instruction,
                                                                continueButtonLabel: instructionScreen.continueButtonLabel ?? "")
                screenConfiguration = VerifaiSingleInstructionScreenConfiguration.customLocal(screenValues: values)
            case .hidden:
                screenConfiguration = VerifaiSingleInstructionScreenConfiguration.hidden
            case .defaultType:
                screenConfiguration = VerifaiSingleInstructionScreenConfiguration.defaultMode
            case .web:
                // Create Web based instruction screen values
                if let urlString = instructionScreen.url, let url = URL(string: urlString) {
                    let values = try VerifaiWebInstructionScreenValues(title: instructionScreen.title ?? "",
                                                                       url: url,
                                                                       continueButtonLabel: instructionScreen.continueButtonLabel ?? "",
                                                                       loader: .webView)
                    screenConfiguration = VerifaiSingleInstructionScreenConfiguration.customWeb(screenValues: values)
                }
            @unknown default:
                // Unknown type
                break
            }
            guard let screenConfiguration = screenConfiguration else { return }
            instructionScreens[instructionScreenEnum] = screenConfiguration
        })
        // Return the processed values
        return try VerifaiInstructionScreenConfiguration(showInstructionScreens: showInstructionScreens,
                                                         instructionScreens: instructionScreens)
    }
    
    /// Convert the scan help configuration coming from Flutter, into one Verifai can understand
    /// - Parameter configuration: The scan help configuration coming from Flutter
    /// - Returns: Verifai's version of the scan help configuration
    private static func verifaiScanHelpConfiguration(from configuration: ScanHelpConfiguration?) -> VerifaiScanHelpConfiguration {
        guard let configuration = configuration else { return VerifaiScanHelpConfiguration() }
        let instruction = NSAttributedString(string: configuration.customScanHelpScreenInstructions ?? "")
        return VerifaiScanHelpConfiguration(isScanHelpEnabled: configuration.showScanHelp?.boolValue ?? true,
                                            customScanHelpScreenInstructions: instruction,
                                            customScanHelpScreenMp4FileName: configuration.customScanHelpScreenMp4FileName ?? "")
    }
    
    // MARK: - Result makers
    
    /// Make a result object that Flutter can understand from the `VerifaiResult` object
    /// - Parameter from: The `VerifaiResult` object from a successful scan
    /// - Returns: A Flutter core result object
    private static func makeCoreResult(from result: VerifaiResult) throws -> CoreResult {
        // Front image and position
        let flutterFrontImage = UIImage.flutterImage(from: result.frontImage)
        let flutterFrontPosition = Polygon.from(result.frontDocumentPosition)
        // Back image and position
        let flutterBackImage = UIImage.flutterImage(from: result.backImage)
        let flutterBackPosition = Polygon.from(result.backDocumentPosition)
        // Document data
        let mrzData = try VerifaiMRZModel.flutterMRZData(from: result.mrzData)
        let document = flutterDocument(from: result.idModel)
        let barcodeResult = flutterBarcodes(from: result.documentBarcodes)
        // Combine the visual inspection results into one dictionary, keeping the keys of the front
        let combinedVisualInspectionResult = result.frontVisualInspectionZoneResult?
            .merging(result.backVisualInspectionZoneResult ?? [:],
                     uniquingKeysWith: {_ , current in current })
        // Create a result Flutter can understand
        return CoreResult.make(withFrontImage: flutterFrontImage,
                               frontDocumentPosition: flutterFrontPosition,
                               backImage: flutterBackImage,
                               backDocumentPosition: flutterBackPosition,
                               mrzData: mrzData,
                               document: document,
                               documentBarcodes: barcodeResult,
                               visualInspectionZoneResult: combinedVisualInspectionResult)
    }
    
    /// Map the values of Verifai's `VerifaiDocumentResult` to Flutter's `DocumentResult`
    /// - Parameter documentResult: The document result object coming from Verifai
    /// - Returns: If it exists an object Flutter can understand
    private static func flutterDocument(from documentResult: VerifaiDocumentResult?) -> DocumentResult? {
        guard let documentResult = documentResult else { return nil }
        // Define holders for data that gets computed
        let numberOfSides = documentResult.zones.contains(where: { $0.side == "BACK" }) ? 2 : 1
        let hasMRZ = documentResult.zones.contains(where: { $0.title == "MRZ" })
        // Build the document zones in a format Flutter understands
        let documentZones = documentResult.zones.map { zone in
            return DocumentZoneResult.make(withTitle: zone.title,
                                           side: zone.side,
                                           x: NSNumber(value: zone.x),
                                           y: NSNumber(value: zone.y),
                                           width: NSNumber(value: zone.width),
                                           height: NSNumber(value: zone.height),
                                           block: NSNumber(value: zone.block))
        }
        // Now we have all the info we need to map the DocumentResult object
        return DocumentResult.make(withUuid: documentResult.uuid,
                                   type: documentResult.type,
                                   model: documentResult.model,
                                   country: documentResult.country,
                                   widthMm: NSNumber(value: documentResult.widthMm),
                                   heightMm: NSNumber(value: documentResult.heightMm),
                                   sampleFront: documentResult.sampleFront,
                                   mrzType: documentResult.mrzType,
                                   mrzStart: documentResult.mrzStart,
                                   nfcType: documentResult.nfcType,
                                   zones: documentZones,
                                   numberOfSides: NSNumber(value: numberOfSides),
                                   hasMrz: NSNumber(value: hasMRZ))
    }
    
    /// Turn Verifai's barcode result array into an array Flutter can understand
    /// - Parameter barcodeResult: The barcode result object coming from Verifai
    /// - Returns: The Barcode result object going to Flutter
    private static func flutterBarcodes(from barcodeResult: [VerifaiBarcodeResult]?) -> [BarcodeResult]? {
        // Ensure there's a result, otherwise no conversion needed
        guard let barcodeResult = barcodeResult else { return nil }
        return barcodeResult.compactMap { result in
            var symbology: BarcodeSymbology
            switch result.symbology {
            case .qr:
                symbology = .qr
            case .code39:
                symbology = .code39
            case .code128:
                symbology = .code128
            case .pdf417:
                symbology = .pdf417
            @unknown default:
                return nil
            }
            return BarcodeResult.make(with: symbology,
                                      side: result.side == .front ? .front : .back,
                                      data: result.data)
        }
    }
    
}
