//
//  Liveness.swift
//  verifai_sdk
//
//  Created by Richard Chirino on 01/04/2022.
//

import Foundation
import VerifaiLivenessKit

class Liveness {
    // Start the Liveness module
    static func start(rootViewController: FlutterViewController?,
                      configuration: LivenessConfiguration,
                      resultHandler: LivenessResultApi?) {
        // Ensure we have a view controller
        guard let rootViewController = rootViewController else { return }
        // Process configuration
        let url: URL? = URL(string: configuration.resultOutputDirectory ?? "")
        let showDismissButton = configuration.showDismissButton?.boolValue ?? true
        let customDismissButtonTitle = configuration.customDismissButtonTitle
        let requirements = verifaiLivenessChecks(from: configuration.checks)
        do {
            // Start the liveness check
            try VerifaiLiveness.start(over: rootViewController,
                                      requirements: requirements,
                                      resultOutputDirectory: url,
                                      showDismissButton: showDismissButton,
                                      customDismissButtonTitle: customDismissButtonTitle) { livenessResult in
                switch livenessResult {
                    
                case .success(let result):
                    do {
                        // Map the result to something flutter can understand and return it
                        let flutterResult = try flutterResult(from: result)
                        resultHandler?.onSuccessResult(flutterResult, completion: { _ in })
                    } catch {
                        resultHandler?.onErrorMessage("🚫 Error in Liveness result: \(error)", completion: { _ in })
                    }
                case .failure(let error):
#if DEBUG
                    print("🚫 Error in Liveness result: \(error)")
#endif
                    resultHandler?.onErrorMessage("🚫 Error in Liveness result: \(error)", completion: { _ in })
                }
            }
        } catch {
            resultHandler?.onErrorMessage("🚫 Error starting Liveness module: \(error)", completion: { _ in })
        }
    }
    
    /// Map the array of liveness checks coming from Flutter into an Array Verifai can understand
    /// - Parameter checks: The checks coming from Flutter
    /// - Returns: An array of checks Verifai can understand
    static private func verifaiLivenessChecks(from checks: [LivenessCheck]?) -> [VerifaiLivenessCheck] {
        // Ensure checks were provided
        guard let checks = checks else {
            // Return the default list of requirements
            return VerifaiLiveness.defaultRequirements(for: Locale.current)
        }
        // Define a holder for the results and populate it
        var result: [VerifaiLivenessCheck] = []
        checks.forEach { check in
            switch check.type {
            case .closeEyes:
                if let verifaiCheck = eyesClosedCheck(args: check.args) {
                    result.append(verifaiCheck)
                }
            case .tilt:
                if let tiltCheck = headTiltCheck(args: check.args) {
                    result.append(tiltCheck)
                }
            case .speech:
                if let speechRecognitionCheck = speechRecognitionCheck(args: check.args) {
                    result.append(speechRecognitionCheck)
                }
            case .faceMatching:
                if let faceMatchCheck = faceMatchCheck(args: check.args) {
                    result.append(faceMatchCheck)
                }
            @unknown default:
                break
            }
        }
        return result
    }
    
    // MARK: - Check creators
    
    /// Create a Verifai eyes closed liveness check
    /// - Parameter args: The arguments required to build the check from the Flutter side
    /// - Returns: If the right arguments were found an eyes closed Liveness check, otherwise nil
    private static func eyesClosedCheck(args: [String: Any]?) -> VerifaiEyesLivenessCheck? {
        // Get duration and optional custom instruction from the args
        if let eyesClosedDuration = args?["numberOfSeconds"] as? Int {
            let instruction = args?["instruction"] as? String ?? nil
            return VerifaiEyesLivenessCheck(duration: TimeInterval(eyesClosedDuration),
                                            instruction: instruction)
        }
        return nil
    }
    
    /// Create a Verifai head tilt liveness check
    /// - Parameter args: The arguments required to build the check from the Flutter side
    /// - Returns: If the right arguments were found a head tilt Liveness check, otherwise nil
    private static func headTiltCheck(args: [String: Any]?) -> VerifaiTiltLivenessCheck? {
        // Get duration and optional custom instruction from the args
        if let faceAngleRequirement = args?["faceAngleRequirement"] as? Float {
            let instruction = args?["instruction"] as? String ?? "Please tilt your head until you reach the green line"
            return VerifaiTiltLivenessCheck(faceAngle: faceAngleRequirement,
                                            instruction: instruction)
        }
        return nil
    }
    
    /// Create a Verifai speech recognition liveness check
    /// - Parameter args: The arguments required to build the check from the Flutter side
    /// - Returns: If the right arguments were found a speech recognition Liveness check, otherwise nil
    private static func speechRecognitionCheck(args: [String: Any]?) -> VerifaiSpeechLivenessCheck? {
        // Get duration and optional custom instruction from the args
        if let speechRequirement = args?["speechRequirement"] as? String {
            let locale = args?["locale"] as? String ?? Locale.current.identifier
            let instruction = args?["instruction"] as? String ?? nil
            return VerifaiSpeechLivenessCheck(speechRequirement: speechRequirement,
                                              locale: Locale(identifier: locale),
                                              instruction: instruction)
        }
        return nil
    }
    
    /// Create a Verifai face match liveness check
    /// - Parameter args: The arguments required to build the check from the Flutter side
    /// - Returns: If the right arguments were found a face match Liveness check, otherwise nil
    private static func faceMatchCheck(args: [String: Any]?) -> VerifaiFaceMatchingLivenessCheck? {
        // Get duration and optional custom instruction from the args
        if let documentImage = args?["documentImage"] as? FlutterStandardTypedData,
           let image = UIImage(data: documentImage.data) {
            let instruction = args?["instruction"] as? String ?? nil
            return VerifaiFaceMatchingLivenessCheck(documentImage: image,
                                                    instruction: instruction)
        }
        return nil
    }
    
    // MARK: - Result converters
    
    private static func flutterResult(from livenessResult: VerifaiLivenessCheckResults) throws -> LivenessResult {
        // Get the computed values
        let withAutomaticChecksPassed = NSNumber(value: livenessResult.automaticChecksPassed)
        let successRatio = NSNumber(value: livenessResult.successRatio)
        // Define a flutter type list holder and get the result list
        var resultList: [LivenessCheckResult] = []
        try livenessResult.resultList.forEach { result in
            // Get general values first
            let status = LivenessCheckStatus.from(result.status)
            let instruction = result.check.instruction
            var checkName = ""
            var checkSpecificData: [String: String] = [:]
            // Build the individual result names
            // Eyes closed
            if let check = result.check as? VerifaiEyesLivenessCheck {
                checkName = "ClosedEyes"
                checkSpecificData = try check.asDictionary()
            }
            // HeadTilt
            if let check = result.check as? VerifaiTiltLivenessCheck {
                checkName = "HeadTilt"
                checkSpecificData = try check.asDictionary()
            }
            // Speech
            if let check = result.check as? VerifaiSpeechLivenessCheck {
                checkName = "Speech"
                checkSpecificData = try check.asDictionary()
            }
            // FaceMatch
            if let check = result.check as? VerifaiFaceMatchingLivenessCheck {
                checkName = "FaceMatchting"
                checkSpecificData = try check.asDictionary()
            }
            // Build the complete result
            let checkResult = LivenessCheckResult.make(withName: checkName,
                                                       instruction: instruction,
                                                       status: status,
                                                       recordingFileName: result.resultUrl?.absoluteString,
                                                       checkSpecificData: checkSpecificData)
            resultList.append(checkResult)
        }
        // We have everything we need to build the result
        return LivenessResult.make(withAutomaticChecksPassed: withAutomaticChecksPassed,
                                   successRatio: successRatio,
                                   resultList: resultList)
    }
    
}
