//
//  NFC.swift
//  verifai_sdk
//
//  Created by Richard Chirino on 28/03/2022.
//

import Foundation
import VerifaiNFCKit
import VerifaiCommonsKit
import OpenSSL

/// Class that handles the communication between the Flutter plugin and the Verifai SDK's NFC module
class NFC {
    // Start the NFC module
    static func start(rootViewController: FlutterViewController?,
                      configuration: NfcConfiguration,
                      resultHandler: NfcResultApi?) {
        // Ensure we have a view controller
        guard let rootViewController = rootViewController,
              let result = Core.sharedResult else { return }
        // Get the configuration values
        let retrieveImage = configuration.retrieveImage?.boolValue ?? false
        let showDismissButton = configuration.showDismissButton?.boolValue ?? false
        let customDismissButtonTitle = configuration.customDismissButtonTitle ?? ""
        let scanHelpConfiguration = verifaiScanHelpConfiguration(from: configuration.scanHelpConfiguration)
        do {
            let instructionScreenConfiguration = try verifaiInstructionScreenConfiguration(from: configuration.instructionScreenConfiguration)
            // Start the NFC
            try VerifaiNFC.start(over: rootViewController,
                                 documentData: result,
                                 retrieveImage: retrieveImage,
                                 showDismissButton: showDismissButton,
                                 customDismissButtonTitle: customDismissButtonTitle,
                                 instructionScreenConfiguration: instructionScreenConfiguration,
                                 scanHelpConfiguration: scanHelpConfiguration,
                                 resultBlock: { nfcResult in
                switch nfcResult {
                case .success(let successResult):
                    do {
                        // Return the result
                        let result = try makeNFCResult(from: successResult)
                        resultHandler?.onSuccessResult(result, completion: { _ in })
                    } catch {
                        resultHandler?.onErrorMessage("🚫 Error creating NFC result: \(error)", completion: { _ in })
                    }
                case .failure(let error):
#if DEBUG
                    print("🚫 Error when performing NFC: \(error)")
#endif
                    resultHandler?.onErrorMessage("🚫 Error when performing NFC: \(error)", completion: { _ in })
                }
            })
        } catch {
            resultHandler?.onErrorMessage("🚫 Error starting NFC module: \(error)", completion: { _ in })
        }
    }
    
    // MARK: - Configuration mappers
    
    /// Map the Flutter version of the instruction screen configuration object to Verifai's native one
    /// - Parameter configuration: The instruction screen configuration coming from flutter
    /// - Returns: Verifai's version of the instruction screen configuration
    private static func verifaiInstructionScreenConfiguration(from configuration: NfcInstructionScreenConfiguration?) throws -> VerifaiNFCInstructionScreenConfiguration {
        // Ensure a configuration was supplied otherwise return default
        guard let configuration = configuration else {
            return try VerifaiNFCInstructionScreenConfiguration(showInstructionScreens: true)
        }
        // Declare a holder for the native results
        var instructionScreens: [VerifaiNFCInstructionScreen: VerifaiNFCSingleInstructionScreenConfiguration] = [:]
        // Get the global values first
        let showInstructionScreens = configuration.showInstructionScreens?.boolValue ?? true
        try configuration.instructionScreens?.forEach({ instructionScreen in
            // NFC currently only has one instruction screen, verify it
            guard instructionScreen.screen == .nfcScanFlowInstruction else { return }
            let instructionScreenEnum = VerifaiNFCInstructionScreen.nfcScanFlowInstruction
            // Declare holder for configuration
            var screenConfiguration: VerifaiNFCSingleInstructionScreenConfiguration?
            // Build the configuration according to the type
            switch instructionScreen.type {
            case .media:
                // Get the values
                let instruction = NSAttributedString(string: instructionScreen.instruction ?? "")
                let values = try VerifaiInstructionScreenValues(title: instructionScreen.title ?? "",
                                                                header: instructionScreen.header ?? "",
                                                                mp4FileName: instructionScreen.mp4FileName ?? "",
                                                                instruction: instruction,
                                                                continueButtonLabel: instructionScreen.continueButtonLabel ?? "")
                screenConfiguration = VerifaiNFCSingleInstructionScreenConfiguration.customLocal(screenValues: values)
            case .hidden:
                screenConfiguration = VerifaiNFCSingleInstructionScreenConfiguration.hidden
            case .defaultType:
                screenConfiguration = VerifaiNFCSingleInstructionScreenConfiguration.defaultMode
            case .web:
                // Create Web based instruction screen values
                if let urlString = instructionScreen.url, let url = URL(string: urlString) {
                    let values = try VerifaiWebInstructionScreenValues(title: instructionScreen.title ?? "",
                                                                       url: url,
                                                                       continueButtonLabel: instructionScreen.continueButtonLabel ?? "",
                                                                       loader: .webView)
                    screenConfiguration = VerifaiNFCSingleInstructionScreenConfiguration.customWeb(screenValues: values)
                }
            @unknown default:
                // Unknown type
                break
            }
            guard let screenConfiguration = screenConfiguration else { return }
            instructionScreens[instructionScreenEnum] = screenConfiguration
        })
        // Return the processed values
        return try VerifaiNFCInstructionScreenConfiguration(showInstructionScreens: showInstructionScreens,
                                                            instructionScreens: instructionScreens)
    }
    
    /// Convert the scan help configuration coming from Flutter, into one Verifai can understand
    /// - Parameter configuration: The scan help configuration coming from Flutter
    /// - Returns: Verifai's version of the scan help configuration
    private static func verifaiScanHelpConfiguration(from configuration: ScanHelpConfiguration?) -> VerifaiScanHelpConfiguration {
        guard let configuration = configuration else { return VerifaiScanHelpConfiguration() }
        let instruction = NSAttributedString(string: configuration.customScanHelpScreenInstructions ?? "")
        return VerifaiScanHelpConfiguration(isScanHelpEnabled: configuration.showScanHelp?.boolValue ?? true,
                                            customScanHelpScreenInstructions: instruction,
                                            customScanHelpScreenMp4FileName: configuration.customScanHelpScreenMp4FileName ?? "")
    }
    
    // MARK: - Result mapping
    
    /// Make a Flutter `NFCResult` object based on Verifai's result object
    /// - Parameter result: The result object coming from an NFC scan in Verifai
    /// - Returns: A Flutter compatible `NFCResult` object
    private static func makeNFCResult(from result: VerifaiNFCResult) throws -> NfcResult {
        // Create needed value holders
        var comSodMatch: Bool? = nil
        // Check document specific data
        if let documentSpecificData = result.documentSpecificData {
            switch documentSpecificData {
            case .dutchEdl(let edlData):
                comSodMatch = edlData.comSodMatch
            default:
                break
            }
        }
        // Values that can be created easily are defined next
        let sodData = result.sodData == nil ? nil : FlutterStandardTypedData(bytes: result.sodData ?? Data())
        // Return values
        return NfcResult.make(withOriginality: NSNumber(value: result.originality),
                              authenticity: NSNumber(value: result.authenticity),
                              confidentiality: NSNumber(value: result.confidentiality),
                              mrzMatch: NSNumber(value: result.mrzMatch),
                              comSodMatch: NSNumber.optional(value: comSodMatch),
                              bacStatus: BacStatus.from(result.bacStatus),
                              activeAuthenticationStatus: AaStatus.from(result.activeAuthenticationStatus),
                              documentCertificateValid: NSNumber(value: result.documentCertificateValid),
                              signingCertificateMatchesWithParent: RootCertificateStatus.from(result.signingCertificateMatchesWithParent),
                              scanCompleted: NSNumber(value: result.scanCompleted),
                              chipAuthenticationStatus: CaStatus.from(result.chipAuthenticationStatus),
                              documentSignatureCorrect: NSNumber(value: result.documentSignatureCorrect),
                              photo: UIImage.flutterImage(from: result.photo),
                              mrzData: try VerifaiMRZModel.flutterMRZData(from: result.mrzData),
                              dataGroups: flutterDataGroupResult(from: result.dataGroups,
                                                                 dataGroupHashes: result.dataGroupHashes),
                              documentCertificate: nil, // Not available on iOS SDK
                              iOSDocumentCertificate: flutterCertificateData(from: result.documentCertificate),
                              sodHashes: flutterSodHashes(from: result.sodHashes),
                              sodData: sodData,
                              featurePoints: nil, // Not available on iOS SDK
                              mimeBytes: nil, // Not available on iOS SDK
                              aaDigestAlgorithm: result.aaDigestAlgorithm,
                              chipAuthenticationOid: result.chipAuthenticationOid,
                              chipAuthenticationAgreementAlgorithm: nil, // Not available on iOS SDK
                              chipAuthenticationPublicKeyAlgorithm: nil, // Not available on iOS SDK
                              signing: nil, // Not available on iOS SDK
                              iOSSigningCertificate: flutterCertificateData(from: result.signingCertificate))
    }
    
    /// Create a Flutter object for the data group and datagroup hashes result coming from Verifai
    /// - Parameters:
    ///   - dataGroups: The data group data in the Verifai result
    ///   - dataGroupHashes: The data group hashes in the Verifai result
    /// - Returns: A result object containing the data group info that Flutter can understand
    private static func flutterDataGroupResult(from dataGroups: [UInt8: Data]?,
                                               dataGroupHashes: [UInt8: Data]?) -> [NSNumber: LdsFile]? {
        // Ensure there is a result
        guard let dataGroups = dataGroups, let dataGroupHashes = dataGroupHashes else { return nil }
        var result: [NSNumber: LdsFile] = [:]
        // Define keys that help us determine read status
        let inaccessibleKey = "Inaccessible".data(using: .utf8)
        dataGroups.forEach { dataGroup in
            // Get the hash
            let hash = dataGroupHashes[dataGroup.key] ?? Data()
            // Determine read status
            let readStatus = dataGroup.value == inaccessibleKey ? ReadStatus.noAccess : ReadStatus.success
            // Create entry and add it
            let ldsFile = LdsFile.make(with: FlutterStandardTypedData(bytes: dataGroup.value),
                                       dataHash: FlutterStandardTypedData(bytes: hash),
                                       readStatus: readStatus)
            result[NSNumber(value: dataGroup.key)] = ldsFile
        }
        return result
    }
    
    /// Return a Flutter data object for the certificates coming from Verifai. The data is in the ASN.1 standard
    /// - Parameter certificate: The certificate coming from Verifai as a hexadecimal string
    /// - Returns: A Flutter data object with the same data
    private static func flutterCertificateData(from certificate: String?) -> FlutterStandardTypedData? {
        // Ensure there's a value
        guard let certificate = certificate else { return nil }
        // Convert the hexadecimal data string coming from verifai to a Flutter data object
        guard let certificateData = certificate.dataFromHexadecimalString() else { return nil }
        return FlutterStandardTypedData(bytes: certificateData)
    }
    
    /// Create a Flutter SodHashes object from the sodHashes data in Verifai
    /// - Parameter hashes: The data coming from Verifai
    /// - Returns: The sod hashes in an object Flutter can understand
    private static func flutterSodHashes(from hashes: [UInt8: Data]?) -> [NSNumber : FlutterStandardTypedData]? {
        guard let hashes = hashes else { return nil }
        // Define a holder
        var result: [NSNumber : FlutterStandardTypedData] = [:]
        // Fill it in with the hash data from Verifai
        hashes.forEach { hash in
            result[NSNumber(value: hash.key)] = FlutterStandardTypedData(bytes: hash.value)
        }
        return result
    }
}
