//
//  Extensions.swift
//  Pods-Runner
//
//  Created by Richard Chirino on 24/03/2022.
//

import Foundation
import VerifaiKit
import VerifaiNFCKit
import VerifaiCommonsKit
import VerifaiLivenessKit

// MARK: - NSNumber
extension NSNumber {
    /// Helper function to transform optional `Bool` into optional `NSNumber`
    /// - Parameter value: The optional `Bool` value that needs to be converted
    /// - Returns: The `NSNumber` representation of the value
    class func optional(value: Bool?) -> NSNumber? {
        guard let value = value else { return nil }
        return NSNumber(value: value)
    }
}

// MARK: - Encodable
extension Encodable {
    /// Helper function to transform encodable objects into dictionary objects that Flutter can understand
    /// - Returns: Dictionary with the values of the encodable object
    func asDictionary() throws -> [String: String] {
        let encoder = JSONEncoder()
        // Goal is to transform the codable object into JSON data and then use native iOS
        // conversion to Dictionary
        let data = try encoder.encode(self)
        guard let dictionary = try JSONSerialization.jsonObject(with: data,
                                                                options: .fragmentsAllowed) as? [String: String] else { return ["Error":"encoding failed"] }
        return dictionary
    }
}

// MARK: - String
extension String {
    func dataFromHexadecimalString() -> Data? {
        let trimmedString = self.trimmingCharacters(
            in: CharacterSet(charactersIn: "<> ")).replacingOccurrences(
                of: " ", with: "")
        
        // Make sure the cleaned up string consists solely of hex digits,
        // and that we have even number of them
        
        let regex = try! NSRegularExpression(pattern: "^[0-9a-f]*$", options: .caseInsensitive)
        
        let found = regex.firstMatch(in: trimmedString, options: [],
                                     range: NSRange(location: 0,
                                                    length: trimmedString.count))
        guard found != nil &&
            found?.range.location != NSNotFound &&
            trimmedString.count % 2 == 0 else {
                return nil
        }
        
        // Everything ok, so now let's build Data
        
        var data = Data(capacity: trimmedString.count / 2)
        var index: String.Index? = trimmedString.startIndex
        
        while let i = index {
            let byteString = String(trimmedString[i ..< trimmedString.index(i, offsetBy: 2)])
            let num = UInt8(byteString.withCString { strtoul($0, nil, 16) })
            data.append([num] as [UInt8], count: 1)
            
            index = trimmedString.index(i, offsetBy: 2, limitedBy: trimmedString.endIndex)
            if index == trimmedString.endIndex { break }
        }
        
        return data
    }
}

// MARK: - Result related extensions
extension UIImage {
    /// Create a Flutter compatible image from an UIImage
    /// - Parameter from: The UIImage that needs to be converted
    /// - Returns: A `FlutterStandardTypedData` object or nil if no image
    static func flutterImage(from image: UIImage?) -> FlutterStandardTypedData? {
        guard let image = image else { return nil }
        // Turn the image into a flutter image
        return FlutterStandardTypedData(bytes: image.pngData() ?? Data())
    }
}

extension VerifaiMRZModel {
    /// Create a Flutter `MRZDataResult` object from a Verifai native one
    /// - Parameter mrzData: The mrzData result from Verifai
    /// - Returns: A Flutter `MRZDataResult` object
    static func flutterMRZData(from mrzData: VerifaiMRZModel?) throws -> MrzDataResult? {
        guard let mrzData = mrzData else { return nil }
        // Map the data over to the Flutter MRZ data object type
        return MrzDataResult.make(withMrzString: mrzData.mrzString,
                                  format: mrzData.format.rawValue,
                                  documentType: mrzData.documentType,
                                  documentSubType: mrzData.documentSubType,
                                  countryCode: mrzData.countryCode,
                                  lastName: mrzData.lastName,
                                  firstName: mrzData.firstName,
                                  documentNumber: mrzData.documentNumber,
                                  documentNumberCheckDigit: mrzData.documentNumberCheckDigit,
                                  nationality: mrzData.nationality,
                                  dateOfBirthRaw: mrzData.dateOfBirth,
                                  dateOfBirthCheckDigit: mrzData.dateOfBirthCheckDigit,
                                  sex: mrzData.sex,
                                  dateOfExpiryRaw: mrzData.dateOfExpiry,
                                  dateOfExpiryCheckDigit: mrzData.dateOfExpiryCheckDigit,
                                  optionalData: mrzData.optionalData,
                                  optionalDataCheckDigit: mrzData.optionalDataCheckDigit,
                                  optionalDataTwo: mrzData.optionalDataTwo,
                                  compositeCheckDigit: mrzData.compositeCheckDigit,
                                  documentSpecificData: try mrzData.documentSpecificData.asDictionary(),
                                  nfcKey: mrzData.nfcKey,
                                  validityScore: NSNumber(value: mrzData.validityScore),
                                  isNfcKeyValid: NSNumber.optional(value: mrzData.isNfcKeyValid),
                                  isDocumentCodeValid: NSNumber(value: mrzData.isDocumentCodeValid),
                                  isDocumentNumberValid: NSNumber.optional(value: mrzData.isDocumentNumberValid),
                                  isDateOfBirthValid: NSNumber.optional(value: mrzData.isDateOfBirthValid),
                                  isDateOfExpiryValid: NSNumber.optional(value: mrzData.isDateOfExpiryValid),
                                  isCompositeValid: NSNumber.optional(value: mrzData.isCompositeValid),
                                  isOptionalDataValid: NSNumber.optional(value: mrzData.isOptionalDataValid),
                                  isLine1LengthValid: NSNumber.optional(value: mrzData.isLine1LengthValid),
                                  isLine2LengthValid: NSNumber.optional(value: mrzData.isLine2LengthValid),
                                  isLine3LengthValid: NSNumber.optional(value: mrzData.isLine3LengthValid))
    }
}

extension VerifaiInstructionScreen {
    /// Get the Verifai instruction screen enum from the Flutter one
    /// - Parameter instructionScreen: The instruction screen enum from Flutter
    /// - Returns: The Verifai version of the instruction screen enum
    static func from(_ instructionScreen: InstructionScreenId) -> VerifaiInstructionScreen {
        switch instructionScreen {
        case .documentChooserTabBarInstruction:
            return .documentChooserTabBarInstruction
        case .mrzPresentFlowInstruction:
            return .mrzPresentFlowInstruction
        case .mrzScanFlowInstruction:
            return .mrzScanFlowInstruction
        @unknown default:
            fatalError("🚫 Unsupported instruction screen type")
        }
    }
}

extension VerifaiDocumentType {
    /// Map Flutter document types (Int based) to Verifai Document types (String based)
    /// - Parameter documentTypes: Document types array coming from Flutter
    /// - Returns: An array of document types in Verifai format
    static func fromArray(_ documentTypes: [Int]) -> [VerifaiDocumentType] {
        return documentTypes.map { documentTypeIndex in
            let documentType = DocumentType(rawValue: UInt(documentTypeIndex))
            switch documentType {
            case .idCard:
                return .idCard
            case .driversLicence:
                return .driversLicence
            case .passport:
                return .passport
            case .refugee:
                return .refugee
            case .emergencyPassport:
                return .emergencyPassport
            case .residencePermitTypeI:
                return .residencePermitTypeI
            case .residencePermitTypeII:
                return .residencePermitTypeII
            case .visa:
                return .visa
            case .unknown, .none, .some:
                return .unknown
            }
        }
    }
}

extension Polygon {
    /// Create a Flutter compatible `Polygon` from a `CGRect` item
    /// - Parameter from: The `CGRect` object to create the `Polygon` from
    /// - Returns: A Flutter compatible Polygon item or nil if the `CGRect` was nil
    static func from(_ rect: CGRect?) -> Polygon? {
        guard let rect = rect else { return nil }
        // Create Flutter type Polygon
        return Polygon.make(withTopLeft: CoordinatePoint.makeWith(x: NSNumber(value: Int(rect.minX)),
                                                                  y: NSNumber(value: Int(rect.maxY))),
                            topRight: CoordinatePoint.makeWith(x: NSNumber(value: Int(rect.maxX)),
                                                               y: NSNumber(value: Int(rect.maxY))),
                            bottomRight: CoordinatePoint.makeWith(x: NSNumber(value: Int(rect.maxX)),
                                                                  y: NSNumber(value: Int(rect.minY))),
                            bottomLeft: CoordinatePoint.makeWith(x: NSNumber(value: Int(rect.minX)),
                                                                 y: NSNumber(value: Int(rect.minY))))
    }
}

// MARK: - NFC Verifai to Flutter return value conversions
extension BacStatus {
    /// Create a Flutter compatible BacStatus object from the one from Verifai
    /// - Parameter verifaiStatus: The BacStatus coming from Verifai
    /// - Returns: The Flutter compatible BacStatus object
    static func from(_ verifaiStatus: VerifaiBacStatus) -> BacStatus {
        switch verifaiStatus {
        case .success:
            return .success
        case .failed:
            return .failed
        case .notChecked:
            return .notChecked
        case .randomFailed:
            return .randomFailed
        @unknown default:
            return .notChecked
        }
    }
}

extension AaStatus {
    /// Create a Flutter compatible AaStatus object from the one from Verifai
    /// - Parameter verifaiStatus: The AaStatus coming from Verifai
    /// - Returns: The Flutter compatible AaStatus object
    static func from(_ verifaiStatus: VerifaiActiveAuthStatus) -> AaStatus {
        switch verifaiStatus {
        case .success:
            return .success
        case .failed:
            return .failed
        case .notChecked:
            return .notChecked
        case .signatureFailed:
            return .signatureFailed
        case .notPresent:
            return .notPresent
        @unknown default:
            return .notChecked
        }
    }
}

extension RootCertificateStatus {
    /// Create a Flutter compatible RootCertificateStatus object from the one from Verifai
    /// - Parameter verifaiStatus: The RootCertificateStatus coming from Verifai
    /// - Returns: The Flutter compatible RootCertificateStatus object
    static func from(_ verifaiStatus: VerifaiRootCertificateStatus) -> RootCertificateStatus {
        switch verifaiStatus {
        case .match:
            return .valid
        case .noMatch:
            return .invalid
        case .revoked:
            return .revoked
        case .notChecked:
            return .notChecked
        case .notFound:
            return .notFound
        @unknown default:
            return .notChecked
        }
    }
}

extension CaStatus {
    /// Create a Flutter compatible CaStatus object from the one from Verifai
    /// - Parameter verifaiStatus: The CaStatus coming from Verifai
    /// - Returns: The Flutter compatible CaStatus object
    static func from(_ verifaiStatus: VerifaiChipAuthStatus) -> CaStatus {
        switch verifaiStatus {
        case .success:
            return .success
        case .failed:
            return .failed
        case .notPresent:
            return .notPresent
        case .notChecked:
            return .notChecked
        @unknown default:
            return .notChecked
        }
    }
}

extension LivenessCheckStatus {
    
    /// Map a liveness check status from Verifai into something flutter can understand
    /// - Parameter verifaiStatus: The liveness check status coming from the Verifai result
    /// - Returns: A liveness check status that Flutter can understand
    static func from(_ verifaiStatus: VerifaiLivenessCheckResult.Status) -> LivenessCheckStatus {
        switch verifaiStatus {
        case .success:
            return .success
        case .undetermined:
            return .undetermined
        case .skipped:
            return .skipped
        case .failed:
            return .failed
        @unknown default:
            return .undetermined
        }
    }
}
